"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.toSwagger = void 0;
const path_to_regexp_1 = require("path-to-regexp");
const generate_schema_1 = __importDefault(require("generate-schema"));
const swagger = {
    openapi: '3.1.0',
    info: {},
    servers: [],
    paths: {},
};
function toSwagger(apidocJson, projectJson) {
    swagger.info = addInfo(projectJson);
    swagger.servers = addServers(projectJson);
    swagger.paths = extractPaths(apidocJson);
    return swagger;
}
exports.toSwagger = toSwagger;
// Removes <p> </p> tags from text
function removeTags(text) {
    return typeof text === 'string' ? text.replace(/(<([^>]+)>)/gi, '') : '';
}
function addInfo(projectJson) {
    return {
        title: projectJson.title || projectJson.name,
        version: projectJson.version,
        description: projectJson.description,
    };
}
function addServers(projectJson) {
    var _a;
    return [
        {
            url: (_a = projectJson.url) !== null && _a !== void 0 ? _a : 'http://localhost/',
            description: 'API Server URL',
        },
    ];
}
/**
 * Extracts paths provided in json format
 * post, patch, put request parameters are extracted in body
 * get and delete are extracted to path parameters
 * @param apidocJson
 * @returns {{}}
 */
function extractPaths(apidocJson) {
    const apiPaths = groupByUrl(apidocJson);
    const paths = {};
    for (const apiPath of apiPaths) {
        const verbs = apiPath.verbs;
        let url = verbs[0].url;
        const pattern = (0, path_to_regexp_1.pathToRegexp)(url);
        const matches = pattern.exec(url);
        // Surrounds URL parameters with curly brackets -> :email with {email}
        for (let j = 1; matches && j < matches.length; j++) {
            const key = matches[j].slice(1);
            url = url.replace(matches[j], `{${key}}`);
        }
        for (const verb of verbs) {
            const obj = (paths[url] = paths[url] || {});
            Object.assign(obj, generateProps(verb));
        }
    }
    return paths;
}
function mapHeaderItem(i) {
    return {
        in: 'header',
        name: i.field,
        description: removeTags(i.description),
        required: !i.optional,
        schema: {
            type: 'string',
            default: i.defaultValue,
            example: i.defaultValue,
        },
    };
}
function mapParameterItem(i) {
    return {
        in: 'path',
        name: i.field,
        description: removeTags(i.description),
        required: !i.optional,
        schema: {
            type: 'string',
            default: i.defaultValue,
            example: i.defaultValue,
        },
    };
}
function mapQueryItem(i) {
    return {
        in: 'query',
        name: i.field,
        description: removeTags(i.description),
        required: !i.optional,
        schema: {
            type: 'string',
            default: i.defaultValue,
            example: i.defaultValue,
        },
    };
}
/**
 * apiDocParams
 * @param {type} type
 * @param {boolean} optional
 * @param {string} field
 * @param {string} defaultValue
 * @param {string} description
 */
/**
 *
 * @param {ApidocParameter[]} apiDocParams
 * @param {Object} parameterInBody
 */
function transferApidocParamsToSwaggerBody(apiDocParams, parameterInBody) {
    const mountPlaces = {
        '': parameterInBody['schema'],
    };
    for (const i of apiDocParams) {
        const type = i.type ? i.type.toLowerCase() : 'string';
        const key = i.field;
        const { objectName = '', propertyName } = createNestedName(i.field);
        if (type.endsWith('object[]')) {
            // if schema(parsed from example) doesn't has this constructure, init
            if (!mountPlaces[objectName]['properties'][propertyName]) {
                mountPlaces[objectName]['properties'][propertyName] = { type: 'array', items: { type: 'object', properties: {}, required: [] } };
            }
            // new mount point
            mountPlaces[key] = mountPlaces[objectName]['properties'][propertyName]['items'];
        }
        else if (type.endsWith('[]')) {
            // if schema(parsed from example) doesn't has this constructure, init
            if (!mountPlaces[objectName]['properties'][propertyName]) {
                mountPlaces[objectName]['properties'][propertyName] = {
                    items: {
                        type: type.slice(0, -2),
                        description: removeTags(i.description),
                        default: i.defaultValue,
                        example: i.defaultValue,
                    },
                    type: 'array',
                };
            }
        }
        else if (type === 'object') {
            // if schema(parsed from example) doesn't has this constructure, init
            if (!mountPlaces[objectName]['properties'][propertyName]) {
                mountPlaces[objectName]['properties'][propertyName] = { type: 'object', properties: {}, required: [] };
            }
            // new mount point
            mountPlaces[key] = mountPlaces[objectName]['properties'][propertyName];
        }
        else {
            mountPlaces[objectName]['properties'][propertyName] = {
                type,
                description: removeTags(i.description),
                default: i.defaultValue,
                example: i.defaultValue,
            };
        }
        if (!i.optional) {
            // generate-schema forget init [required]
            if (mountPlaces[objectName]['required']) {
                mountPlaces[objectName]['required'].push(propertyName);
            }
            else {
                mountPlaces[objectName]['required'] = [propertyName];
            }
        }
    }
    return parameterInBody;
}
function generateProps(verb) {
    const pathItemObject = {};
    const parameters = generateParameters(verb);
    const body = generateBody(verb);
    const responses = generateResponses(verb);
    const type = verb.type.toLowerCase();
    pathItemObject[type] = {
        tags: [verb.group],
        summary: removeTags(verb.name),
        description: removeTags(verb.title),
        requestBody: { content: { 'application/json': body } },
        parameters,
        responses,
    };
    return pathItemObject;
}
function generateBody(verb) {
    var _a;
    const mixedBody = [];
    if ((_a = verb === null || verb === void 0 ? void 0 : verb.parameter) === null || _a === void 0 ? void 0 : _a.fields) {
        mixedBody.push(...(verb.parameter.fields.Body || []));
    }
    if (verb === null || verb === void 0 ? void 0 : verb.body) {
        mixedBody.push(...(verb.body || []));
    }
    let body = {};
    if (mixedBody.length) {
        body = generateRequestBody(verb, mixedBody);
    }
    return body;
}
function generateParameters(verb) {
    var _a, _b, _c;
    let parameter = [];
    let query = [];
    const header = ((_b = (_a = verb === null || verb === void 0 ? void 0 : verb.header) === null || _a === void 0 ? void 0 : _a.fields) === null || _b === void 0 ? void 0 : _b.Header) || [];
    if ((_c = verb === null || verb === void 0 ? void 0 : verb.parameter) === null || _c === void 0 ? void 0 : _c.fields) {
        parameter = verb.parameter.fields.Parameter || [];
        query = verb.parameter.fields.Query || [];
    }
    const parameters = [];
    parameters.push(...parameter.map(mapParameterItem));
    parameters.push(...query.map(mapQueryItem));
    parameters.push(...header.map(mapHeaderItem));
    parameters.push(...(verb.query || []).map(mapQueryItem));
    return parameters;
}
function generateRequestBody(verb, mixedBody) {
    var _a, _b, _c;
    const bodyParameter = {
        schema: {
            description: '',
            properties: {},
            type: 'object',
            required: [],
        },
    };
    if (((_b = (_a = verb === null || verb === void 0 ? void 0 : verb.parameter) === null || _a === void 0 ? void 0 : _a.examples) === null || _b === void 0 ? void 0 : _b.length) > 0) {
        const example = verb.parameter.examples[0];
        const { json } = safeParseJson(example.content);
        bodyParameter.schema = generate_schema_1.default.json((_c = example.title) !== null && _c !== void 0 ? _c : '', json);
        //bodyParameter.description = example.title;
    }
    transferApidocParamsToSwaggerBody(mixedBody, bodyParameter);
    return bodyParameter;
}
function generateResponses(verb) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
    const success = verb.success;
    const error = verb.error;
    const responses = {
        200: {
            description: 'OK',
        },
    };
    if (((_a = success === null || success === void 0 ? void 0 : success.examples) === null || _a === void 0 ? void 0 : _a.length) > 0) {
        for (const example of success.examples) {
            const { code, json } = safeParseJson(example.content, 200);
            const schema = generate_schema_1.default.json(example.title, json);
            responses[code] = {
                description: (_b = example.title) !== null && _b !== void 0 ? _b : 'OK',
                content: {
                    'application/json': {
                        schema: schema,
                        examples: Object.assign(Object.assign({}, ((_e = (_d = (_c = responses[code]) === null || _c === void 0 ? void 0 : _c.content) === null || _d === void 0 ? void 0 : _d.examples) !== null && _e !== void 0 ? _e : {})), { [example.title]: { value: json } }),
                    },
                },
            };
        }
    }
    if (((_f = error === null || error === void 0 ? void 0 : error.examples) === null || _f === void 0 ? void 0 : _f.length) > 0) {
        for (const example of error.examples) {
            const { code, json } = safeParseJson(example.content, 400);
            const schema = generate_schema_1.default.json(example.title, json);
            responses[code] = {
                description: (_g = example.title) !== null && _g !== void 0 ? _g : 'ERROR',
                content: {
                    'application/json': {
                        schema: schema,
                        examples: Object.assign(Object.assign({}, ((_k = (_j = (_h = responses[code]) === null || _h === void 0 ? void 0 : _h.content) === null || _j === void 0 ? void 0 : _j.examples) !== null && _k !== void 0 ? _k : {})), { [example.title]: { value: json } }),
                    },
                },
            };
        }
    }
    mountResponseSpecSchema(verb, responses);
    return responses;
}
function mountResponseSpecSchema(verb, responses) {
    var _a, _b, _c, _d, _e, _f, _g, _h;
    let fields = {};
    if ((_a = verb === null || verb === void 0 ? void 0 : verb.success) === null || _a === void 0 ? void 0 : _a.fields) {
        fields = Object.assign(Object.assign({}, fields), verb.success.fields);
    }
    if ((_b = verb === null || verb === void 0 ? void 0 : verb.error) === null || _b === void 0 ? void 0 : _b.fields) {
        fields = Object.assign(Object.assign({}, fields), verb.error.fields);
    }
    for (const field in fields) {
        const responseCode = parseInt((_d = (_c = field.match(/\b([1-5][0-9]{2})\b/)) === null || _c === void 0 ? void 0 : _c[0]) !== null && _d !== void 0 ? _d : '');
        if (isNaN(responseCode)) {
            continue;
        }
        const responseSpec = fields[field];
        const responseExamples = (_g = (_f = (_e = responses[responseCode]) === null || _e === void 0 ? void 0 : _e.content) === null || _f === void 0 ? void 0 : _f['application/json']) === null || _g === void 0 ? void 0 : _g.examples;
        const bodyParameter = {
            schema: {
                properties: {},
                type: 'object',
                required: [],
            },
        };
        transferApidocParamsToSwaggerBody(responseSpec, bodyParameter);
        responses[responseCode] = {
            description: (_h = responseSpec.title) !== null && _h !== void 0 ? _h : '',
            content: {
                'application/json': Object.assign(Object.assign({}, bodyParameter), { examples: responseExamples }),
            },
        };
    }
}
function safeParseJson(content, defaultCode = 200) {
    // such as  'HTTP/1.1 200 OK\n' +  '{\n' + ...
    let startingIndex = 0;
    for (let i = 0; i < content.length; i++) {
        const character = content[i];
        if (character === '{' || character === '[') {
            startingIndex = i;
            break;
        }
    }
    const mayCodeString = content.slice(0, startingIndex);
    const mayContentString = content.slice(startingIndex);
    const mayCodeSplit = mayCodeString.trim().split(' ');
    const code = mayCodeSplit.length === 3 ? parseInt(mayCodeSplit[1]) : defaultCode;
    let json = {};
    try {
        json = JSON.parse(mayContentString.replace(/,([ \t\n]+[}\])])/g, '$1'));
    }
    catch (error) {
        const message = mayContentString + (error instanceof Error ? error.message : '');
        console.warn('\x1b[0;33mparse error\x1b[0m: ' + message);
    }
    return { code, json };
}
function createNestedName(field, defaultObjectName) {
    let propertyName = field;
    let objectName;
    const propertyNames = field === null || field === void 0 ? void 0 : field.split('.');
    if (propertyNames && propertyNames.length > 1) {
        propertyName = propertyNames.pop();
        objectName = propertyNames.join('.');
    }
    return {
        propertyName: propertyName,
        objectName: objectName || defaultObjectName,
    };
}
function groupByUrl(apidocJson) {
    const grouped = Object.entries(apidocJson.reduce((groups, item) => {
        const key = item['url'];
        if (!groups[key]) {
            groups[key] = [];
        }
        groups[key].push(item);
        return groups;
    }, {}));
    const results = grouped.map(([url, verbs]) => ({ url, verbs }));
    return results;
}
